package de.tu_darmstadt.tk.SmartHomeNetworkSim.control;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Protocol;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;

/**
 * Controller for manipulation of the network model with methods like  
 * 
 *
 * @author Andreas T. Meyer-Berg
 */
public class NetworkController {

	/**
	 * Model which will be manipulated
	 */
	private Model model;
	/**
	 * Controller which can be use
	 */
	private Controller controller;
	
	/**
	 * Creates a new NetworkController, which may manipulate the given model and use the controller
	 * @param model
	 * @param controller
	 */
	public NetworkController(Model model, Controller controller) {
		this.model = model;
		this.controller = controller;
	}

	/**
	 * Validate all device positions, move all devices into the bounds, if they are outside the visualization
	 */
	public void validateDevicePosition() {
		// Update all device positions
		for (SmartDevice d : model.getDevices()) {
			d.setX(controller.getSettingsController().scalePos(d.getX(), 1.0, controller.getSettingsController().getWidth()));
			d.setY(controller.getSettingsController().scalePos(d.getY(), 1.0, controller.getSettingsController().getHeight()));
			d.setZ(controller.getSettingsController().scalePos(d.getZ(), 1.0, controller.getSettingsController().getDepth()));
		}
	}

	/**
	 * Adds SmartDevice to the Model and verifies that it is inside the model
	 * bounds
	 * 
	 * @param sd
	 *            SmartDevice which should be added to the model
	 */
	public void addSmartDevice(SmartDevice sd) {
		if (model.getDevices().contains(sd))
			return;
		model.addDevices(sd);
		// validate Position
		sd.setX(controller.getSettingsController().scalePos(sd.getX(), 1.0, controller.getSettingsController().getWidth()));
		sd.setY(controller.getSettingsController().scalePos(sd.getY(), 1.0, controller.getSettingsController().getHeight()));
		sd.setZ(controller.getSettingsController().scalePos(sd.getZ(), 1.0, controller.getSettingsController().getDepth()));
	}

	/**
	 * Creates a new SmartDevice at the given Position. The Device is moved
	 * into the model bounds, if the position is outside the bounds
	 * 
	 * @param name
	 *            name of the smartDevice
	 * @param x_pos
	 *            position on the x-Axis
	 * @param y_pos
	 *            position on the y-Axis
	 * @param z_pos
	 *            position on the z-Axis
	 */
	public void createSmartDevice(String name, int x_pos, int y_pos, int z_pos) {
		SmartDevice sd = new SmartDevice(name);
		sd.setX(controller.getSettingsController().scalePos(x_pos, 1.0, controller.getSettingsController().getWidth()));
		sd.setY(controller.getSettingsController().scalePos(y_pos, 1.0, controller.getSettingsController().getHeight()));
		sd.setZ(controller.getSettingsController().scalePos(z_pos, 1.0, controller.getSettingsController().getDepth()));
		model.addDevices(sd);
	}

	/**
	 * Deletes the given SmartDevice toDelete, removes it from its links and
	 * connections.
	 * 
	 * @param toDelete
	 *            the SmartDevice to delete
	 */
	public void deleteSmartDevice(SmartDevice toDelete) {
		if (toDelete == null)
			return;
	
		// Delete Connections
		for (Port p : toDelete.getPorts()) {
			//Skip ports that are not connected
			if(p.getConnection()==null)continue;
			removeDeviceFromConnection(p, p.getConnection());
			
			/**
			 * Connection of the current port
			 *
			Connection c = p.getConnection();
			
			//remove from protocol
			if(c.getProtocol()!=null)
				c.getProtocol().removeDevice(p);
			
			//remove from connection
			c.removeSmartDevice(p);
			if (c.getStatus()==Connection.FINISHED||c.getStatus()==Connection.TERMINATED) {
				//if connection terminated - remove rest of ports
				for (Port sd : c.getParticipants()) {
					if (sd != null && sd != p)
						sd.getOwner().removePort(sd);
				}
			}*/
		}
		
		// Remove from Links
		LinkedList<Link> links = new LinkedList<Link>(toDelete.getLinks());
		for (Link link : links)
			removeSmartDeviceFromLink(toDelete, link);
		links.clear();
		// Remove Links from Device
		toDelete.getLinks().clear();
				
		//Remove all ports from the device
		toDelete.getPorts().clear();
		model.getDevices().remove(toDelete);
	}

	/**
	 * Removes the SmartDevice from the given Link
	 * 
	 * @param toRemove the Device that should be removed from the link
	 * @param link the Link, which contains the SmartDevice
	 */
	public void removeSmartDeviceFromLink(SmartDevice toRemove, Link link) {
		if(link != null){
			link.removeDevice(toRemove);
			if(link.getDevices().size()==0)
				deleteLink(link);
		}
		if(toRemove!=null)
			toRemove.removeLink(link);
	}

	/**
	 * Moves the SmartDevice device to the specified location, if it exists
	 * 
	 * @param device
	 *            the device to move
	 * @param x
	 *            new x position
	 * @param y
	 *            new y position
	 * @param z
	 *            new z position
	 */
	public void moveSmartDevice(SmartDevice device, int x, int y, int z) {
		device.setX(x);
		device.setY(y);
		device.setZ(z);
	
	}

	/**
	 * Adds Link to the model
	 * @param link link to add
	 */
	public void addLink(Link link){
		model.addConnectionNetwork(link);
	}

	/**
	 * Removes Link from the Model
	 * @param link link to remove
	 */
	public void removeLink(Link link){
		model.getConnectionNetworks().remove(link);
	}

	public Collection<Link> getLinks(){
		return model.getConnectionNetworks();
	}

	public void addConnection(Connection connection){
		model.addConnection(connection);
	}

	public void removeConnection(Connection connection){
		model.getConnections().remove(connection);
	}

	public Collection<Connection> getConnections(){
		return model.getConnections();
	}

	public void addLinkToDevice(Link newLink, SmartDevice smartDevice) {
		if(!newLink.getDevices().contains(smartDevice)){
			newLink.addDevice(smartDevice);
			smartDevice.addLink(newLink);
		}
		
	}

	public void removeLinkFromDevice(Link link, SmartDevice smartDevice) {
		if(smartDevice!=null)
			smartDevice.removeLink(link);
		if(link!=null){
			link.removeDevice(smartDevice);
			if(link.getDevices().size()==0)
				deleteLink(link);
		}
	}

	/**
	 * Changes Roles of the device to the newRole, returns true, if the role was successfully changed, false if not.
	 * If false is returned, the device will no longer be part of this connection.
	 * 
	 * @param protocol protocol which should be edited
	 * @param con Connection, which contains the protocol
	 * @param device device which should be added
	 * @param newRole new role of the device
	 * @return true if new role was set, false if not
	 */
	public boolean changeRoleOfDevice(Protocol protocol, Connection con, Port device, int newRole){
		if(newRole < 0 || newRole >= protocol.getNumberOfRoles()){
			protocol.removeDevice(device);
			removeDeviceFromConnection(device, con);
			return false;
		} else if(protocol.getDevicesWithRole(newRole).contains(device)){
			if(!con.getParticipants().contains(device))
				con.addSmartDevice(device);
			if(device.getConnection()!=con)
				device.setConnection(con);
			return true;
		}
		else{
			protocol.removeDevice(device);
			boolean added = protocol.addDeviceOfRole(device, newRole);
			if(added){
				if(!con.getParticipants().contains(device))
					con.addSmartDevice(device);
				if(device.getConnection()!=con)
					device.setConnection(con);
				return true;
			}else{
				if(con.getParticipants().contains(device))
					con.removeSmartDevice(device);
				if(device.getConnection()!=null)
					device.setConnection(null);
				return false;
			}
		}
	}

	/**
	 * Removes Device p from the Connection
	 * @param p Port/Device to remove
	 * @param connection connection, which should remove the device
	 */
	public void removeDeviceFromConnection(Port p, Connection connection){
		if(connection != null){
			connection.removeSmartDevice(p);
			connection.getProtocol().removeDevice(p);
			//TODO: Protocol ?
			if(connection.getProtocol()!=null)
				connection.getProtocol().removeDevice(p);
			if(connection.getParticipants().size() == 0)
				deleteConnection(connection);
		}
		if(p!=null)
			p.setConnection(null);
	}

	/**
	 * Adds Device p to the connection with the specified role
	 * @param p Port/Device to be added
	 * @param connection Connection
	 * @param role new role of the device (See {@link Protocol} Roles)
	 * @return true, if the device was added
	 */
	public boolean addDeviceToConnection(Port p, Connection connection, int role){
		if(connection.getProtocol().getDevicesWithRole(role).contains(p) || connection.getProtocol().addDeviceOfRole(p, role)){
			//If port already has the role, or it could be assigned - just check the fields
			if(!connection.getParticipants().contains(p))
				connection.addSmartDevice(p);
			if(p.getConnection()!=connection)
				p.setConnection(connection);
			return true;
		}else {
			//Device could not be added -> Remove
			if(p.getConnection()==connection)
				p.setConnection(null);
			if(connection.getParticipants().contains(p))
				connection.removeSmartDevice(p);
			if(connection.getProtocol().getDevices().contains(p))
				connection.getProtocol().removeDevice(p);
			return false;
		}
	}

	/**
	 * Changes the link of the given connection to the new link. Returns true if it was successfully changed, false if it could not be changed.
	 * @param connection
	 * @param newLink
	 * @return true on successful change, false on failure & restore
	 */
	public boolean changeLinkOfConnection(Connection connection, Link newLink) {
		if(newLink != null){
			connection.getLink().removeConnection(connection);
			connection.setLink(newLink);
			for(Port p:connection.getParticipants()){
				if(p.getOwner()!=null&& !newLink.getDevices().contains(p.getOwner())){
					addLinkToDevice(newLink, p.getOwner());
				}
			}
			connection.setLink(newLink);
			newLink.addConnection(connection);
			return true;
		}else{
			return false;
		}
	}

	/**
	 * Changes the type of the connection to the new Type and updates all references
	 * @param connection connection to be updated
	 * @param newConnectionClass class of new type
	 * @return newly created connection
	 */
	public Connection changeConnectionType(Connection connection, Class<? extends Connection> newConnectionClass) {
		
		Connection newCon = null;
		try{
			newCon = newConnectionClass.newInstance();
			newCon.setProtocol(connection.getProtocol());
			newCon.setLink(connection.getLink());
			connection.getLink().addConnection(newCon);
			newCon.setStatus(connection.getStatus());
			newCon.setPacketLossProbability(connection.getPacketLossProbability());
			newCon.setName(connection.getName());
			for (Iterator<Port> p = connection.getParticipants().iterator(); p.hasNext();) {
				Port type = (Port) p.next();
				newCon.addSmartDevice(type);
				type.setConnection(newCon);
				
			}
		}catch(Exception e){
			System.out.println("Error while changing protocol: "+e.toString());
			/**
			 * Restore old connection on Failure
			 */
			if(newCon != null){
				newCon.setProtocol(null);
				if(newCon.getLink()!=null)
					newCon.getLink().removeConnection(newCon);
				newCon.setLink(null);
				for (Iterator<Port> p = connection.getParticipants().iterator(); p.hasNext();) {
					Port type = (Port) p.next();
					if(type.getConnection()!=connection)
						type.setConnection(connection);
					newCon.removeSmartDevice(type);
					
				}
				
				
			}
			return null;
		}
		if(getConnections().contains(connection)){
			removeConnection(connection);
		}
		addConnection(newCon);	
		controller.getSettingsController().getConfigurationManager().getSelectionModel().clickedConnection.clear();
			
			
		connection.setLink(null);
		connection.setProtocol(null);
		connection.setStatus(Connection.TERMINATED);
		connection.setName("deleted");
		for(Port p: newCon.getParticipants()){
			connection.removeSmartDevice(p);			
		}
		
		controller.notifyObservers();
		
		return newCon;
	}

	/**
	 * Deletes the network model, removes all Devices, Connections and Links
	 */
	public void deleteNetworkModel(){
		/**
		 * Devices which should be deleted
		 */
		LinkedList<SmartDevice> devicesToDelete = new LinkedList<SmartDevice>(model.getDevices());
		for(SmartDevice d: devicesToDelete)
			deleteSmartDevice(d);
		devicesToDelete.clear();
		/**
		 * Connections which should be deleted
		 */
		LinkedList<Connection> connectionsToDelete = new LinkedList<Connection>(model.getConnections());
		for(Connection c: connectionsToDelete)
			deleteConnection(c);
		connectionsToDelete.clear();
		/**
		 * Links which should be deleted
		 */
		LinkedList<Link> linksToDelete = new LinkedList<Link>(model.getConnectionNetworks());
		for(Link l: model.getConnectionNetworks())
			deleteLink(l);
		linksToDelete.clear();
		/**
		 * Update the GUI
		 */
		controller.notifyObservers();
	}

	/**
	 * Deletes the Connection c and all references
	 * @param c Connection to be deleted
	 */
	public void deleteConnection(Connection c) {
		if(c == null)return;
		LinkedList<Port> ports = new LinkedList<Port>(c.getParticipants());
		for(Port p:ports)
			removeDeviceFromConnection(p, c);
		ports.clear();
		removeConnectionFromLink(c, c.getLink());
		c.setStatus(Connection.TERMINATED);
		removeConnection(c);
	}

	/**
	 * Deletes Link l and all references
	 * @param l Link to be deleted
	 */
	public void deleteLink(Link l) {
		if(l==null)return;
		LinkedList<SmartDevice> devices = new LinkedList<SmartDevice>(l.getDevices());
		for(SmartDevice d : devices)
			removeLinkFromDevice(l, d);
		devices.clear();
		LinkedList<Connection> connections = new LinkedList<Connection>(l.getConnections());
		for(Connection c:connections)
			removeConnectionFromLink(c,l);
		connections.clear();
		l.getPackets().clear();
		removeLink(l);
	}

	/**
	 * Removes Connection from Link
	 * @param c Connection to be removed
	 * @param l Link to be removed
	 */
	public void removeConnectionFromLink(Connection c, Link l) {
		if(c!=null && c.getLink()==l){
			c.setLink(null);
		}
		if(l!=null){
			l.removeConnection(c);		
		}
	}

	/**
	 * Changes the type of the link to new type, specified by the given Link-class
	 * @param oldLink oldLink, whose attributes will b copied to the new Link 
	 * @param newType Type/Class of the new Link
	 * @return newly created Link, null on failure/error
	 */
	public Link changeLinkType(Link oldLink, Class<? extends Link> newType) {
		if(newType == null)
			return null;
		/**
		 * New Link which was created
		 */
		Link newLink = null;
		try{
			newLink = newType.newInstance();
		}catch(Exception e){
			return null;
		}
		if (newLink == null) {
			return null;
		}else {
			// Set old Name
			newLink.setName(oldLink.getName());
			// Add to Mode
			if(getLinks().contains(oldLink)){
				removeLink(oldLink);
				addLink(newLink);
			}
			// Add devices to the new Link
			LinkedList<SmartDevice> devices= new LinkedList<>(oldLink.getDevices());
			for(SmartDevice device:devices){
				removeLinkFromDevice(oldLink, device);
				addLinkToDevice(newLink, device);
			}
		return newLink;
		}
	}

}
